/* lab10a.c */

#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>

#define I_MAX_THREADS   254
#define I_MAX_PAGINAS   100
#define SZ_FORMATO      "Formato: p <#threads> <#paginas>\n"
#define SZ_MAX_THREADS  "O maximo de threads permitido e' %d\n"
#define SZ_MAX_PAGINAS  "O maximo de paginas permitido e' %d\n"
#define SZ_POS_PAGINAS  "O total de paginas deve ser positivo\n"
#define SZ_ERRO_CRIA    "Erro na criacao do thread %d\n"
#define SZ_NUM_THREADS  "Numero de threads %d\n"
#define SZ_TOT_PAGINAS  "Total de paginas %d\n"
#define SZ_PEDINDO      "Processo %d pedindo %d pginas\n"
#define SZ_CONSEGUE     "Processo %d conseguiu alocar %d pginas\n"
#define SZ_LIBERA       "Processo %d liberou %d pginas\n"


typedef struct {
   int iNumPaginas ,
       iIdThread   ;
} tpPares ;

typedef struct {
   pthread_mutex_t MonitorPaginas         ;
   pthread_cond_t  turn[ I_MAX_THREADS ]  ;          
   int             iTotPaginas            ,
                   iNumPares              ;
   tpPares         pairs[ I_MAX_THREADS ] ;    
} tpMonPaginas ;


int          n           ,
             iMaxPaginas ;
tpMonPaginas MonPaginas  ; 


void* UsarPaginas( void* MeuId ) ;

void request( int amount ,
              int iMeuId  ) ;

void release( int amount ,
              int iMeuId  ) ;

void Sinalizar( void ) ;

void InserirPar( int iNumPaginas ,
                 int iMeuId       ) ;

int iRemoverPar( void ) ;

void Esperar( int iMeuId ) ;


void* UsarPaginas( void* MeuId ) {
   int iNumPaginas                  ,
       iMeuId      = *( int* )MeuId ;

   srand( time( NULL ) * ( iMeuId + 1 ) ) ;

   while ( 1 ) {
      iNumPaginas = 1 + ( int )( ( ( double )iMaxPaginas * rand( ) )
                               / ( RAND_MAX + 1.0 ) ) ;

      printf( SZ_PEDINDO, iMeuId, iNumPaginas ) ;
      request( iNumPaginas, iMeuId ) ;
      Esperar( iMeuId ) ;
      release( iNumPaginas / 2, iMeuId ) ;
      iNumPaginas -= iNumPaginas / 2 ;
      Esperar( iMeuId ) ;
      release( iNumPaginas, iMeuId ) ;
   } /* while */
} /* UsarPaginas */


void request( int amount ,
              int iMeuId  ) {

   pthread_mutex_lock( &MonPaginas.MonitorPaginas ) ;
   
   while ( amount > MonPaginas.iTotPaginas ) {
      InserirPar( amount, iMeuId ) ;
      pthread_cond_wait( &MonPaginas.turn[ iMeuId ],
                                         &MonPaginas.MonitorPaginas ) ;
   } /* while */

   MonPaginas.iTotPaginas -= amount ;
   printf( SZ_CONSEGUE, iMeuId, amount ) ;
   Sinalizar( ) ;
   
   pthread_mutex_unlock( &MonPaginas.MonitorPaginas ) ;

} /* request */


void release( int amount ,
              int iMeuId  ) {

   pthread_mutex_lock( &MonPaginas.MonitorPaginas ) ;

   MonPaginas.iTotPaginas += amount ;
   printf( SZ_LIBERA, iMeuId, amount ) ;
   Sinalizar( ) ;

   pthread_mutex_unlock( &MonPaginas.MonitorPaginas ) ;

} /* release */


void Sinalizar( void ) {
   int iEscolhido = iRemoverPar( ) ;

   if ( iEscolhido >= 0 ) {
      pthread_cond_signal( &MonPaginas.turn[ iEscolhido ] ) ;
   } /* if */
} /* Sinalizar */


void InserirPar( int iNumPaginas ,
                 int iMeuId       ) {

   MonPaginas.pairs[ MonPaginas.iNumPares ].iNumPaginas = iNumPaginas ;
   MonPaginas.pairs[ MonPaginas.iNumPares ].iIdThread = iMeuId ;
   MonPaginas.iNumPares++ ;

} /* InserirPar */


int iRemoverPar( void ) {
   int i                                               ,
       iPrimeiro   = 0                                 ,
       iIdPrimeiro = MonPaginas.pairs[ 0 ].iIdThread   ,
       iMinPaginas = MonPaginas.pairs[ 0 ].iNumPaginas ;

   for ( i = 1 ; i < MonPaginas.iNumPares ; i++ ) {
      if ( MonPaginas.pairs[ i ].iNumPaginas < iMinPaginas ) {
         iPrimeiro = i ;
         iIdPrimeiro = MonPaginas.pairs[ i ].iIdThread ;
         iMinPaginas = MonPaginas.pairs[ i ].iNumPaginas ;
      } /* if */
   } /* for */

   if ( MonPaginas.iNumPares > 0 
     && iMinPaginas <= MonPaginas.iTotPaginas ) {
      for ( i = iPrimeiro ; i < MonPaginas.iNumPares - 1 ; i++ ) {
         MonPaginas.pairs[ i ].iNumPaginas
                              = MonPaginas.pairs[ i + 1 ].iNumPaginas ;
         MonPaginas.pairs[ i ].iIdThread = MonPaginas.pairs[ i + 1 ].iIdThread ;
      } /* for */

      MonPaginas.iNumPares-- ;
      return iIdPrimeiro ;
   } /* if */

   return -1 ;
} /* iRemoverPar */


void Esperar( int iMeuId ) {
   unsigned long int i                                        ,
                     lTotal = ( unsigned long )10000000L 
                            * ( unsigned long )( iMeuId + 1 ) ;

   for ( i = 0 ; i < lTotal ; i++ ) { }
} /* Esperar */


int main( int   argc    ,
          char* argv[ ]  ) {
   int       i                       ,
             iErro                   ;
   int*      pMeuId                  ;
   pthread_t Thread[ I_MAX_THREADS ] ;

   if ( argc != 3 ) {
      printf( SZ_FORMATO ) ;
      return 0 ;
   } /* if */

   n = atoi( argv[ 1 ] ) ;
   iMaxPaginas = atoi( argv[ 2 ] ) ;

   if ( n > I_MAX_THREADS ) {
      printf( SZ_MAX_THREADS, I_MAX_THREADS ) ;
      return 0 ;
   } /* if */

   if ( iMaxPaginas > I_MAX_PAGINAS ) {
      printf( SZ_MAX_PAGINAS, I_MAX_PAGINAS ) ;
      return 0 ;
   } /* if */

   if ( iMaxPaginas <= 0 ) {
      printf( SZ_POS_PAGINAS ) ;
      return 0 ;
   } /* if */

   MonPaginas.iNumPares = 0 ;
   MonPaginas.iTotPaginas = iMaxPaginas ;
   pthread_mutex_init( &MonPaginas.MonitorPaginas, NULL ) ;
   
   for ( i = 0 ; i < I_MAX_THREADS ; i++ ) {
      pthread_cond_init( &MonPaginas.turn[ i ], NULL ) ;
   } /* for */

   printf( SZ_NUM_THREADS, n ) ;
   printf( SZ_TOT_PAGINAS, iMaxPaginas ) ;

   for ( i = 0 ; i < n ; i++ ) {
      pMeuId = ( int* )malloc( sizeof( int ) ) ;
      *pMeuId = i ;

      iErro = pthread_create( &Thread[ i ], NULL, UsarPaginas,
                                                  ( void* )( pMeuId ) ) ;

      if ( iErro != 0 ) {
         printf( SZ_ERRO_CRIA, i ) ;
         return 1 ;
      } /* if */
   } /* for */

   pthread_exit( NULL ) ;

   return 0 ;
}
